'''### _Subsurface specific containers_

These are containers relevant within subsurface workflows. Most of them
rely on the setting `scratch_ensemble` configuration within the
`container_settings`.
I.e. you could have
```yaml
title: Reek Webviz Demonstration
username: some_username
password: some_password

container_settings:
  scratch_ensembles:
    iter-0: /scratch/my_ensemble/realization-*/iter-0
    iter-1: /scratch/my_ensemble/realization-*/iter-1

pages:

 - title: Front page
   content:
    - container: SummaryStats
      ensemble: iter-0
```
'''

from ._summary_stats import SummaryStats
from ._parameter_distribution import ParameterDistribution
from ._disk_usage import DiskUsage
from ._subsurface_map import SubsurfaceMap
from ._history_match import HistoryMatch
from ._intersect import Intersect


__all__ = ['SummaryStats',
           'ParameterDistribution',
           'DiskUsage',
           'SubsurfaceMap',
           'HistoryMatch',
           'Intersect']



from uuid import uuid4
import pandas as pd
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import cache
from ..datainput import scratch_ensemble


class ParameterDistribution:
    '''### Parameter distribution

This container shows parameter distribution as histogram,
and correlation between the parameters as a correlation matrix.

* `ensemble`: Which ensemble in `container_settings` to visualize.
* `title`: Optional title for the container.
'''

    def __init__(self, app, container_settings, ensemble,
                 title: str = 'Parameter Distribution'):

        self.title = title

        self.dropdown_vector_id = 'dropdown-vector-{}'.format(uuid4())
        self.radio_plot_type_id = 'radio-plot-type-{}'.format(uuid4())
        self.chart_id = 'chart-id-{}'.format(uuid4())
        self.histogram_div_id = 'histogram-div-{}'.format(uuid4())

        # Finding all parameters:
        self.ensemble_path = container_settings['scratch_ensembles'][ensemble]
        self.parameter_columns = sorted(list(
            get_parameters(self.ensemble_path).columns))

        self.set_callbacks(app)

    @property
    def layout(self):
        return html.Div([
            html.H2(self.title),
            html.P('Plot type:', style={'font-weight': 'bold'}),
            dcc.RadioItems(id=self.radio_plot_type_id,
                           options=[{'label': i, 'value': i} for i in
                                    ['Histogram', 'Pairwise correlation']],
                           value='Histogram'),
            html.Div(id=self.histogram_div_id,
                     children=[
                         html.P('Parameter:',
                                style={'font-weight': 'bold'}),
                         dcc.Dropdown(id=self.dropdown_vector_id,
                                      clearable=False,
                                      options=[{'label': i, 'value': i} for
                                               i in self.parameter_columns],
                                      value=self.parameter_columns[0]),
                     ]),
            dcc.Graph(id=self.chart_id,
                      config={
                          'displaylogo': False,
                          'modeBarButtonsToRemove': ['sendDataToCloud']
                      }
                      )
        ])

    def set_callbacks(self, app):
        @app.callback(Output(self.chart_id, 'figure'),
                      [Input(self.dropdown_vector_id, 'value'),
                       Input(self.radio_plot_type_id, 'value')])
        def update_plot(parameter, plot_type):
            if plot_type == 'Histogram':
                return render_histogram(self.ensemble_path, parameter)
            if plot_type == 'Pairwise correlation':
                return render_matrix(self.ensemble_path)

        @app.callback(Output(self.histogram_div_id, 'style'),
                      [Input(self.radio_plot_type_id, 'value')])
        def toggle_parameter_selector(plot_type):
            if plot_type == 'Histogram':
                return {'display': 'block'}
            if plot_type == 'Pairwise correlation':
                return {'display': 'none'}

    def add_webvizstore(self):
        return [(get_parameters, [{'ensemble_path': self.ensemble_path}])]


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def get_parameters(ensemble_path) -> pd.DataFrame:
    ens = scratch_ensemble('', ensemble_path)

    return ens.parameters


@cache.memoize(timeout=cache.TIMEOUT)
def render_histogram(ensemble_path, parameter):
    data = {
        'x': get_parameters(ensemble_path)[parameter],
        'type': 'histogram'
    }

    layout = {
        'bargap': 0.05,
        'font': {'family': 'Equinor'},
        'xaxis': {'family': 'Equinor'},
        'yaxis': {'family': 'Equinor'},
        'hoverlabel': {'font': {'family': 'Equinor'}}
    }

    return {'data': [data], 'layout': layout}


@cache.memoize(timeout=cache.TIMEOUT)
def render_matrix(ensemble_path):

    data = get_parameters(ensemble_path)
    values = list(data.corr().values)

    data = {
        'type': 'heatmap',
        'x': data.columns,
        'y': data.columns,
        'z': values
    }

    layout = {
        'margin': {'l': 200},
        'font': {'family': 'Equinor'},
        'xaxis': {'family': 'Equinor'},
        'yaxis': {'family': 'Equinor'},
        'hoverlabel': {'font': {'family': 'Equinor'}}
    }

    return {'data': [data], 'layout': layout}



from uuid import uuid4
import json
import pandas as pd
import dash_html_components as html
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import cache
from webviz_subsurface_components import Map
from ..datainput import scratch_ensemble


class SubsurfaceMap:
    '''### Subsurface map

This container visualizes the subsurface. Currently only supporting reservoir
model grid maps. In addition to show a map, it can visualize the flow pattern
in the simulation output using streamlines.

* `ensemble`: Which ensemble in `container_settings` to visualize.
* `map_value`: Which property to show in the map (e.g. `PERMX`).
* `flow_value`: Which property to use for the streamlines animation
  (e.g. `FLOWAT`).
* `time_step`: Which report or time step to use in the simulation output.
* `title`: Optional title for the container.
'''

    def __init__(self, container_settings, ensemble, map_value: str,
                 flow_value: str, time_step, title: str = 'Subsurface map'):

        self.title = title
        self.map_id = 'map-{}'.format(uuid4())
        self.map_value = map_value
        self.flow_value = flow_value
        self.time_step = time_step

        self.ensemble_path = container_settings['scratch_ensembles'][ensemble]
        self.map_data = get_map_data(self.ensemble_path, self.map_value,
                                     self.flow_value, self.time_step)

    @property
    def layout(self):
        return html.Div([
                   html.H2(self.title),
                   Map(id=self.map_id, data=self.map_data)
               ])

    def add_webvizstore(self):
        return [(get_uncompressed_data, [{'ensemble_path': self.ensemble_path,
                                          'map_value': self.map_value,
                                          'flow_value': self.flow_value,
                                          'time_step': self.time_step}])]


@cache.memoize(timeout=cache.TIMEOUT)
def get_map_data(ensemble_path, map_value, flow_value,
                 time_step):
    '''Returns map data in the format of a JSON string, suitable for the
    corresponding subsurface map component in
    https://github.com/equinor/webviz-subsurface-components
    '''

    grid = get_uncompressed_data(ensemble_path, map_value, flow_value,
                                 time_step)

    INDICES_COL = ['i', 'j', 'k']
    X_COL = ['x0', 'x1', 'x2', 'x3']
    Y_COL = ['y0', 'y1', 'y2', 'y3']
    FLOW_COL = ['FLOWI+', 'FLOWJ+']

    RESOLUTION = 1000

    grid = grid[INDICES_COL + X_COL + Y_COL + ['value'] + FLOW_COL]
    grid = grid[grid['value'] > 0]

    xmin, xmax = grid[X_COL].values.min(), grid[X_COL].values.max()
    ymin, ymax = grid[Y_COL].values.min(), grid[Y_COL].values.max()

    flowmin, flowmax = grid[FLOW_COL].values.min(), grid[FLOW_COL].values.max()

    valmin, valmax = grid['value'].min(), grid['value'].max()

    if (xmax - xmin) > (ymax - ymin):
        coord_scale = RESOLUTION/(xmax - xmin)
    else:
        coord_scale = RESOLUTION/(ymax - ymin)

    grid[X_COL] = (grid[X_COL] - xmin) * coord_scale
    grid[Y_COL] = (grid[Y_COL] - ymin) * coord_scale
    grid[X_COL + Y_COL] = grid[X_COL + Y_COL].astype(int)

    flow_scale = RESOLUTION/(flowmax - flowmin)
    grid[FLOW_COL] = (grid[FLOW_COL] - flowmin) * flow_scale
    grid[FLOW_COL] = grid[FLOW_COL].astype(int)

    val_scale = RESOLUTION/(valmax - valmin)
    grid['value'] = (grid['value'] - valmin) * val_scale
    grid['value'] = grid['value'].astype(int)

    grid[INDICES_COL] = grid[INDICES_COL].astype(int)

    data = {
            'values': grid.values.tolist(),
            'linearscales': {
                'coord': [float(coord_scale), float(xmin), float(ymin)],
                'value': [float(val_scale), float(valmin)],
                'flow': [float(flow_scale), float(flowmin)]
                            }
            }

    return json.dumps(data, separators=(',', ':'))


@webvizstore
def get_uncompressed_data(ensemble_path, map_value, flow_value,
                          time_step) -> pd.DataFrame:

    ens = scratch_ensemble('', ensemble_path)

    properties = [map_value, f'{flow_value}I+', f'{flow_value}J+']
    if 'PERMX' not in properties:
        properties.append('PERMX')

    grid = ens.get_eclgrid(properties, report=time_step)

    grid = grid[grid['PERMX'] > 0]  # Remove inactive grid cells

    grid['value'] = grid[map_value]
    grid['FLOWI+'] = grid[f'{flow_value}I+']
    grid['FLOWJ+'] = grid[f'{flow_value}J+']

    # Webviz map component uses different corner point terminology than libecl
    for (new, old) in [('x0', 'x1'), ('x1', 'x2'), ('x2', 'x4'),
                       ('y0', 'y1'), ('y1', 'y2'), ('y2', 'y4')]:
        grid[new] = grid[old]

    return grid



import os
import pandas as pd
try:
    import xtgeo
except ImportError:
    pass
from webviz_config.common_cache import cache


@cache.memoize(timeout=cache.TIMEOUT)
def load_well(well_name):
    return xtgeo.well.Well(well_name)


def load_surface(s_name, real_path, surface_cat):
    path = os.path.join(real_path, 'share/results/maps',)
    s_path = os.path.join(path, f'{s_name}--{surface_cat}.gri')
    try:
        return xtgeo.surface.RegularSurface(s_path)
    except IOError:
        raise IOError


@cache.memoize(timeout=cache.TIMEOUT)
def get_wfence(well_name, extend=200, tvdmin=0):
    well = load_well(well_name)
    data = well.get_fence_polyline(sampling=20, extend=extend, tvdmin=tvdmin)
    df = pd.DataFrame(data)
    df.columns = df.columns.astype(str)
    return df


@cache.memoize(timeout=cache.TIMEOUT)
def get_hfence(well, surface):
    return surface.get_fence(get_wfence(well).values.copy())



import mock
import dash
import pandas as pd
from pytest_dash.wait_for import wait_for_element_by_css_selector
from webviz_config.common_cache import cache
from webviz_subsurface.containers import _parameter_distribution


# mocked functions
get_parameters = 'webviz_subsurface.containers'\
                '._parameter_distribution.get_parameters'


def test_parameter_dist(dash_threaded):

    app = dash.Dash(__name__)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True
    app.config.suppress_callback_exceptions = True
    cache.init_app(app.server)
    driver = dash_threaded.driver
    container_settings = {'scratch_ensembles': {'iter-0': ''}}
    ensemble = 'iter-0'

    with mock.patch(get_parameters) as mock_parameters:
        mock_parameters.return_value = pd.read_csv('tests/data/parameters.csv')

        p = _parameter_distribution.\
            ParameterDistribution(app, container_settings, ensemble)

        app.layout = p.layout
        dash_threaded(app)

        my_component = wait_for_element_by_css_selector(
                           driver,
                           f'#{p.dropdown_vector_id}'
                       )

        if 'REAL' != my_component.text:
            raise AssertionError()



from uuid import uuid4
import pandas as pd
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output
from webviz_plotly.graph_objs import FanChart
from webviz_config.webviz_store import webvizstore
from webviz_config.common_cache import cache
from ..datainput import scratch_ensemble


class SummaryStats:
    '''### Summary statistics

This container visualizes simulation profiles, both per realization and
statistical plots (min, max, mean, p10, p90).

* `ensemble`: Which ensemble in `container_settings` to visualize.
* `sampling`: Optional. Either `monthly` or `yearly`. Default is `monthly`.
* `title`: Optional title for the container.
'''

    def __init__(
            self,
            app,
            container_settings,
            ensemble,
            sampling: str = 'monthly',
            title: str = 'Simulation time series'):

        self.title = title
        self.dropwdown_vector_id = 'dropdown-vector-{}'.format(uuid4())
        self.sampling = sampling
        self.radio_plot_type_id = 'radio-plot-type-{}'.format(uuid4())
        self.chart_id = 'chart-id-{}'.format(uuid4())

        # Finding all summary vectors:
        self.ensemble_path = container_settings['scratch_ensembles'][ensemble]

        self.smry_columns = sorted(list(get_summary_data(self.ensemble_path,
                                                         self.sampling)
                                        .drop(columns=['DATE', 'REAL'])
                                        .columns))

        self.set_callbacks(app)

    @property
    def layout(self):
        return html.Div([
            html.H2(self.title),
            html.P('Summary Vector:', style={'font-weight': 'bold'}),
            dcc.Dropdown(id=self.dropwdown_vector_id,
                         clearable=False,
                         options=[{'label': i, 'value': i}
                                  for i in self.smry_columns],
                         value=self.smry_columns[0]),
            html.P('Plot type:', style={'font-weight': 'bold'}),
            dcc.RadioItems(id=self.radio_plot_type_id,
                           options=[{'label': i, 'value': i}
                                    for i in ['Realizations', 'Statistics']],
                           value='Realizations'),
            dcc.Graph(id=self.chart_id,
                      config={
                          'displaylogo': False,
                          'modeBarButtonsToRemove': ['sendDataToCloud']
                      }
                      )
        ])

    def set_callbacks(self, app):
        @app.callback(Output(self.chart_id, 'figure'),
                      [Input(self.dropwdown_vector_id, 'value'),
                       Input(self.radio_plot_type_id, 'value')])
        def update_plot(vector, summary_plot_type):
            if summary_plot_type == 'Realizations':
                return render_realization_plot(
                    self.ensemble_path,
                    self.sampling, vector)
            if summary_plot_type == 'Statistics':
                return render_stat_plot(
                    self.ensemble_path,
                    self.sampling, vector)

    def add_webvizstore(self):
        return [(get_summary_data, [{'ensemble_path': self.ensemble_path,
                                     'sampling': self.sampling,
                                     'statistics': False}]),
                (get_summary_data, [{'ensemble_path': self.ensemble_path,
                                     'sampling': self.sampling,
                                     'statistics': True}])]


@cache.memoize(timeout=cache.TIMEOUT)
@webvizstore
def get_summary_data(ensemble_path, sampling,
                     statistics=False) -> pd.DataFrame:

    ens = scratch_ensemble('', ensemble_path)
    if statistics:
        return ens.get_smry_stats(time_index=sampling)
    else:
        return ens.get_smry(time_index=sampling)


@cache.memoize(timeout=cache.TIMEOUT)
def render_realization_plot(ensemble_path, sampling, vector):

    data = get_summary_data(ensemble_path,
                            sampling)[['REAL', 'DATE', vector]]

    traces = [{
        'x': df['DATE'],
        'customdata': df['REAL'],
        'y': df[vector],
        'name': name,
        'type': 'line'
    } for name, df in data.groupby('REAL') if name != 'DATE']

    layout = {
        'hovermode': 'closest',
        'barmode': 'overlay',
        'bargap': 0.05,
        'xaxis': {'title': 'Date', 'family': 'Equinor'},
        'yaxis': {'title': vector, 'family': 'Equinor'},
        'font': {'family': 'Equinor'},
        'hoverlabel': {'font': {'family': 'Equinor'}},
    }

    return {'data': traces, 'layout': layout}


@cache.memoize(timeout=cache.TIMEOUT)
def render_stat_plot(ensemble_path, sampling, vector):

    data = get_summary_data(ensemble_path, sampling,
                            statistics=True)[vector].unstack().transpose()

    data['name'] = vector
    data.rename(index=str, inplace=True,
                columns={"minimum": "min", "maximum": "max"})

    return FanChart(data.iterrows())
